# coding:utf-8
import typing
from argparse import ZERO_OR_MORE, ONE_OR_MORE
import decimal
import fractions
from urllib.parse import ParseResult as URL
from bourbaki.introspection.types import (
    NonStrCollection,
    is_named_tuple_class,
    get_named_tuple_arg_types,
)
from bourbaki.introspection.generic_dispatch import (
    GenericTypeLevelSingleDispatch,
    UnknownSignature,
)
from .utils import maybe_map
from .exceptions import CLIIOUndefined


NoneType = type(None)

NestedCollectionTypes = (
    typing.Mapping[NonStrCollection, typing.Any],
    typing.Mapping[typing.Any, NonStrCollection],
)


class AmbiguousUnionNargs(CLIIOUndefined):
    def __str__(self):
        return "Types in union {} imply an ambiguous number of command line args".format(
            self.type_
        )


class NestedCollectionsCLIArgError(CLIIOUndefined):
    def __str__(self):
        return (
            "Some type parameters of sequence type {} require more than one command line arg; can't parse "
            "unambiguously".format(self.type_)
        )


def check_union_nargs(*types):
    types = [t for t in types if t not in (NoneType, None)]
    all_nargs = tuple(maybe_map(cli_nargs, types, (UnknownSignature, CLIIOUndefined)))
    if len(set(all_nargs)) > 1:
        raise AmbiguousUnionNargs((typing.Union, *types))
    if len(all_nargs) == 0:
        raise CLIIOUndefined((typing.Union, *types))
    return all_nargs


def check_tuple_nargs(tup_type, *types, allow_tail_collection: bool = True):
    if Ellipsis in types:
        types = [typing.List[types[0]]]
    all_nargs = tuple(cli_nargs(t) for t in types)
    head_nargs = all_nargs[:-1] if allow_tail_collection else all_nargs
    if any(((a is not None) and (not isinstance(a, int))) for a in head_nargs):
        raise NestedCollectionsCLIArgError((tup_type, *types))

    if not all_nargs:
        total_nargs = 0
    elif allow_tail_collection:
        tail_nargs = all_nargs[-1]
        if tail_nargs == ONE_OR_MORE:
            total_nargs = ONE_OR_MORE
        elif tail_nargs == ZERO_OR_MORE:
            total_nargs = ONE_OR_MORE if head_nargs else ZERO_OR_MORE
        else:
            total_nargs = sum(1 if n is None else n for n in all_nargs)
    else:
        total_nargs = sum(1 if n is None else n for n in all_nargs)

    return all_nargs, total_nargs


# nargs for argparse.ArgumentParser

cli_nargs = GenericTypeLevelSingleDispatch("cli_nargs", isolated_bases=[typing.Union])

cli_nargs.register_all(decimal.Decimal, fractions.Fraction, URL, as_const=True)(None)


@cli_nargs.register(typing.Any)
def default_nargs(*args, **kwargs):
    # single string arg unless otherwise overridden below
    return None


@cli_nargs.register(NonStrCollection)
def seq_nargs(*types):
    # all collections other than str
    return ZERO_OR_MORE


@cli_nargs.register_all(NestedCollectionTypes)
def nested_collections_cli_error(t, *args):
    # collections of collections
    raise NestedCollectionsCLIArgError((t, *args))


@cli_nargs.register(typing.Collection[NonStrCollection])
@cli_nargs.register(typing.Tuple[NonStrCollection, ...])
def nested_option_nargs(t, *types):
    if len(types) > 1 and not (len(types) == 2 and types[1] is Ellipsis):
        raise CLIIOUndefined(t, *types)
    return cli_nargs(types[0])


@cli_nargs.register(typing.Tuple)
def tuple_nargs(t, *types):
    if not types and is_named_tuple_class(t):
        types = get_named_tuple_arg_types(t)
    elif types and types[-1] is Ellipsis:
        return cli_nargs(typing.List[types[0]])
    elif not types:
        return ZERO_OR_MORE
    else:
        _, nargs = check_tuple_nargs(t, *types)
        return nargs

    _, nargs = check_tuple_nargs(t, *types)
    return nargs


@cli_nargs.register(typing.Union)
def union_nargs(u, *types):
    all_nargs = check_union_nargs(*types)
    return next(iter(all_nargs))


cli_option_nargs = cli_nargs

cli_action = GenericTypeLevelSingleDispatch(
    "cli_action", isolated_bases=[typing.Union, typing.Tuple]
)


@cli_action.register(typing.Union)
def cli_action_any(u, *ts):
    actions = set(map(cli_action, (t for t in ts if t is not NoneType)))
    if len(actions) > 1:
        raise CLIIOUndefined(u, *ts)
    return next(iter(actions))


cli_action.register(typing.Any, as_const=True)(None)
cli_action.register(typing.Collection[NonStrCollection], as_const=True)("append")



# coding:utf-8
from warnings import warn
from datetime import datetime
from ..reutils import (
    find_lambda,
    sub_lambda,
    extract_groups,
    datetime_regex,
    named_group,
    re_escaped,
)
from .defaults import DEFAULT_LOG_MSG_FMT, DEFAULT_LOG_DATE_FMT
from .regexes import *


def log_fmt_fields(log_fmt: str):
    fields = list(find_lambda(log_fmt_field_pat, log_fmt, extract_groups))
    return fields


def log_line_regex(log_fmt: str, date_fmt: str = None):
    def get_field_re(match):
        name, typ = match.groups()
        if typ != log_fields[name]:
            warn(
                "default printf formatting style for '%s' is '%s' but '%s' is "
                "specified instead" % (name, log_fields[name], typ)
            )
        pat = field_regexes.get(typ, string_field_regexes.get(name))
        if pat is None:
            if name == "asctime":
                pat = datetime_regex(date_fmt)
            else:
                raise ValueError("No pattern available for field name '%s'" % name)

        return named_group(pat, name)

    pat = sub_lambda(log_fmt_field_pat, log_fmt, get_field_re, re_escaped)
    return pat


def lambda_mutate(d, converters):
    for k, f in converters.items():
        if k in d:
            d[k] = f(d[k])
    return d


def join_stacktrace(trace):
    if len(trace) > 0:
        return "".join(trace)
    else:
        return None


def begins_stacktrace(line):
    return stacktrace_start_re.match(line)


def check_message_last(fields):
    not_last = any(t[0] == "message" for t in fields) and fields[-1][0] != "message"
    if not_last:
        warn(
            "The 'message' field of the log format is not placed last; regex parsing "
            "of log files may be slow or impossible"
        )
    return not not_last


def log_file_record_iter(filepath, log_fmt, date_fmt, raise_=False):
    if isinstance(filepath, str):
        file = open(filepath, "r")
        close = True
    else:
        file = filepath
        close = False

    def strptime(s):
        return datetime.strptime(s, date_fmt)

    line_re = re.compile(log_line_regex(log_fmt, date_fmt))

    fields = log_fmt_fields(log_fmt)
    message_last = check_message_last(fields)

    converters = {name: field_converters[typ] for name, typ in fields}
    if "asctime" in converters:
        converters["asctime"] = strptime
    if "message" in converters:
        converters["message"] = join_stacktrace
    converters["stackTrace"] = join_stacktrace

    lines = iter(file)
    last_record = dict(stackTrace=[], message="")
    append_trace = False
    counter = 0

    for line in lines:
        record = line_re.match(line)
        if not record:
            append_trace = append_trace or begins_stacktrace(line)
            if append_trace:
                last_record["stackTrace"].append(line)
            elif message_last:
                last_record["message"] += "\n%s" % line.rstrip("\n")
            else:
                _warn_or_raise(
                    "could not parse line: '{}'".format(line.rstrip("\n")),
                    IOError,
                    raise_,
                )
        else:
            # we have a new record; yield the last one
            if counter > 0:
                yield lambda_mutate(last_record, converters)

            last_record = record.groupdict()
            last_record["stackTrace"] = []
            append_trace = False
        counter += 1

    if counter > 0:
        # there were at least some lines in the file; yield the last parsed
        yield lambda_mutate(last_record, converters)

    if close:
        file.close()


def log_file_to_df(
    filepath,
    log_fmt=DEFAULT_LOG_MSG_FMT,
    date_fmt=DEFAULT_LOG_DATE_FMT,
    datetime_index=False,
    raise_=False,
):
    import pandas

    fields = [t[0] for t in log_fmt_fields(log_fmt)]

    fields.append("stackTrace")

    if datetime_index:
        if "asctime" not in fields:
            raise ValueError(
                "datetime_index is specified but 'asctime' is not one of "
                "the log format fields"
            )

    df = pandas.DataFrame.from_records(
        log_file_record_iter(filepath, log_fmt, date_fmt, raise_), columns=fields
    )

    if datetime_index:
        df.set_index("asctime", inplace=True)
    if not df.index.is_monotonic:
        df.sort_index(inplace=True)

    return df


def _warn_or_raise(errmsg, exception, raise_):
    if raise_:
        raise exception(errmsg)
    else:
        warn(errmsg)
    return



# !/usr/bin/env python
"""
Usage:
    compgen_python_classpaths.py IMPORT_PREFIX [legal_prefix1 [legal_prefix2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --class [legal_prefix1 [legal_prefix2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --callable [legal_prefix1 [legal_prefix2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --subclass [classpath1 [classpath2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --instance [classpath1 [classpath2 [...]]]
"""

from itertools import chain
import os
import sys
import typing
from typing import Union
from warnings import warn

scriptname = "compgen_python_classpaths.py"

Module = type(sys)

issubclass_generic = None
isinstance_generic = None
typetypes = None
nontypes = tuple(
    getattr(typing, t)
    for t in ["Optional", "ClassVar", "NoReturn"]
    if hasattr(typing, t)
)

COMPLETION_DEBUG_ENV_VAR = "BOURBAKI_COMPLETION_DEBUG"

DEBUG = os.environ.get(COMPLETION_DEBUG_ENV_VAR, "").lower().strip()
if DEBUG == "true" or (DEBUG.isdigit() and DEBUG != "0"):
    warn(
        "Found environment variable {}={}; verbose output will be generated for {}".format(
            COMPLETION_DEBUG_ENV_VAR, DEBUG, scriptname
        )
    )
    DEBUG = True if not DEBUG.isdigit() else int(DEBUG)
else:
    DEBUG = False

CLASS_FLAG = "--class"
CALLABLE_FLAG = "--callable"
MODULE_FLAG = "--module"
INSTANCE_FLAG = "--instance"
SUBCLASS_FLAG = "--subclass"


# CLI dispatches to one of these functions


def complete_objpath(prefix: str = "", *legal_prefixes: str):
    _debug("Completing paths for all objects")
    yield from _complete_objpath(prefix, legal_prefixes=legal_prefixes)


def complete_callable_path(prefix: str = "", *legal_prefixes: str):
    _debug("Completing paths for callables")
    yield from _complete_objpath(
        prefix, typecheck=callable_, legal_prefixes=legal_prefixes
    )


def complete_classpath(prefix: str = "", *legal_prefixes: str):
    _debug("Completing paths for classes")
    yield from _complete_objpath(
        prefix, cls=type, typecheck=isinstance, legal_prefixes=legal_prefixes
    )


def complete_module_path(prefix: str = "", *legal_prefixes: str):
    _debug("Completing paths for modules")
    yield from _complete_objpath(
        prefix,
        include_attrs=False,
        legal_prefixes=legal_prefixes,
        include_builtins=False,
    )


def complete_instance_path(prefix: str = "", *legal_classes: str):
    _debug("Completing paths for instances of : {cls}", cls=legal_classes)
    # delayed import for speed if not needed; this only gets called at most once per interpreter launch
    global isinstance_generic
    if isinstance_generic is None:
        from bourbaki.introspection.typechecking import isinstance_generic

    if legal_classes:
        cls = _materialize_classes(*legal_classes)
        typecheck = isinstance
    else:
        cls = typecheck = None

    yield from _complete_objpath(prefix, cls=cls, typecheck=typecheck)


def complete_subclasspath(prefix: str = "", *legal_superclasses: str):
    _debug("Completing paths for subclasses of : {cls}", cls=legal_superclasses)
    # delayed import for speed if not needed; this only gets called at most once per interpreter launch
    global issubclass_generic
    if issubclass_generic is None:
        global typetypes
        from bourbaki.introspection.types import issubclass_generic, typetypes

    if legal_superclasses:
        cls = _materialize_classes(*legal_superclasses)
        typecheck = issubclass_
    else:
        cls, typecheck = type, isinstance

    yield from _complete_objpath(prefix, cls=cls, typecheck=typecheck)


# The above dispatch to this main function


def _complete_objpath(
    prefix: str,
    cls=None,
    typecheck=None,
    legal_prefixes=None,
    include_attrs=True,
    include_builtins=True,
):
    from bourbaki.introspection.imports import import_object
    import builtins

    if legal_prefixes:
        candidate_prefixes = [p for p in legal_prefixes if p.startswith(prefix)]
        if not candidate_prefixes:
            return ()
    else:
        candidate_prefixes = [prefix]

    def inner(prefix):
        parts = prefix.split(".")
        if len(parts) > 1:
            modname, suffix = ".".join(parts[:-1]), parts[-1]
            names = get_all_module_names(prefix, suffix)
            try:
                mod = import_object(modname)
            except ImportError:
                mod = None
            else:
                names = chain(names, get_submodule_names(mod, suffix, modname=modname))
                if include_attrs:
                    names = chain(
                        names,
                        get_attr_names(mod, suffix, cls, typecheck, modname=modname),
                    )
        else:
            mod = None
            names = get_all_module_names(prefix)
            if include_builtins:
                names = chain(names, get_attr_names(builtins, prefix, cls, typecheck))
        return names

    return chain.from_iterable(map(inner, candidate_prefixes))


# Helpers


def _debug(msg: str, **kw):
    if DEBUG:
        OKGREEN = "\033[92m"
        WARNING = "\033[93m"
        FAIL = "\033[91m"
        ENDC = "\033[0m"
        print(
            "\033[92m{}: {}\033[0m".format(scriptname, msg.format(**kw)),
            file=sys.stderr,
        )


def callable_(o, cls=None):
    return callable(o)


def issubclass_(o, cls):
    # delayed import for speed if not needed; this only gets called at most once per interpreter launch
    if isinstance(o, (type, *typetypes)) and o not in nontypes:
        try:
            return issubclass_generic(o, cls)
        except Exception:
            return False
    else:
        return False


def _materialize_classes(*legal_superclasses: str):
    if not legal_superclasses:
        clss = None
    else:
        from bourbaki.introspection.imports import import_type
        from bourbaki.introspection.types import typetypes

        clss = []
        for name in legal_superclasses:
            try:
                cls = import_type(name)
            except (ImportError, AttributeError):
                pass
            else:
                if isinstance(cls, (type, *typetypes)):
                    clss.append(cls)
        clss = tuple(clss)

    _debug("Imported classes: {clss}", clss=clss)
    if not clss:
        return object
    return Union[clss]


def get_attr_names(mod, attr_prefix="", cls=None, typecheck=None, modname=None):
    _debug("Completing attributes on module/object {mod}", mod=mod)

    fmt_name = None if modname is None else (modname + ".{}").format

    attrs = dir(mod)

    if attr_prefix:
        _debug("Filtering to attributes with prefix {prefix}", prefix=attr_prefix)
        attrs = (a for a in attrs if a.startswith(attr_prefix))

    if typecheck is not None:
        _debug(
            "Filtering to attributes by type using {func}({cls})",
            func=typecheck,
            cls=cls,
        )
        attrs = (a for a in attrs if typecheck(getattr(mod, a, None), cls))

    yield from (map(fmt_name, attrs) if fmt_name else attrs)


def get_submodule_names(mod, suffix="", modname=None):
    file_ = getattr(mod, "__file__", None)

    if modname is None:
        modname = getattr(mod, "__name__", "")

    if file_ and os.path.basename(file_) == "__init__.py":
        dirs = getattr(mod, "__path__", [os.path.dirname(file_)])
        for dir_ in dirs:
            _debug(
                "Completing submodules of {modname} in {dir_}",
                modname=modname,
                dir_=dir_,
            )
            for name in get_module_names(dir_, suffix):
                yield ".".join([modname, name])


def get_module_names(dir_, submod_prefix=None):
    # _debug("Searching modules in {dir_}", dir_=dir_)
    for name in os.listdir(dir_):
        if name == "__init__.py":
            continue
        if submod_prefix and not name.startswith(submod_prefix):
            continue

        prefix, ext = os.path.splitext(name)
        if ext == ".py":
            yield prefix
        elif ext == ".pth":
            yield prefix.split("-")[0]
        elif ext == ".so":
            yield prefix.split(".")[0]
        elif ext == ".egg-link":
            yield prefix.replace("-", "_")
        elif ext not in (".dist-info", ".egg-info"):
            d = os.path.join(dir_, name)
            if os.path.isdir(d) and os.path.isfile(os.path.join(d, "__init__.py")):
                yield name


def get_all_module_names(prefix="", submod_prefix=None):
    _debug(
        "Listing all known modules from sys.path{prefix_note}",
        prefix_note="" if not prefix else " with prefix '%s'" % prefix,
    )
    dirs = [d for d in sys.path if os.path.isdir(d)]

    names = (name for dir_ in dirs for name in get_module_names(dir_, submod_prefix))
    names = chain(names, ("builtins",))
    if prefix:
        names = (name for name in names if name.startswith(prefix))

    yield from uniq(names)


def uniq(iter_):
    visited = set()
    for _ in iter_:
        if _ in visited:
            continue
        visited.add(_)
        yield _


def main():
    argv = sys.argv[1:]
    if not argv or argv[0] in ("-h", "--help"):
        print(__doc__)
        exit(0)

    if os.getenv("APPUTILS_COMPLETION_DEBUG", "false").lower() == "true":
        DEBUG = True

    dispatch = {
        CLASS_FLAG: complete_classpath,
        SUBCLASS_FLAG: complete_subclasspath,
        INSTANCE_FLAG: complete_instance_path,
        MODULE_FLAG: complete_module_path,
        CALLABLE_FLAG: complete_callable_path,
    }

    prefix, args = argv[0], argv[1:]

    if not args:
        args = [""]
        complete = complete_objpath
    elif args[0] in dispatch:
        complete = dispatch[args[0]]
        args = args[1:]
    else:
        complete = complete_objpath

    total_completions = 0
    name = None
    for name in complete(prefix, *args):
        print(name, file=sys.stdout)
        total_completions += 1
    if total_completions == 1:
        # prevent early-stopping in the case of only 1 completion that doesn't actually meet the criteria
        for name in complete(name + ".", *args):
            print(name, file=sys.stdout)


if __name__ == "__main__":
    main()



# coding:utf-8
# decorators for functions defined in a class def context to override default behaviors for how they are configured
# as subcommands
from typing import Collection
from bourbaki.introspection.callables import funcname
from .helpers import _maybe_bool, _validate_parse_order

NO_OUTPUT_HANDLER = object()


class cli_spec:
    @staticmethod
    def allow_positional_args(f):
        """override the require_options setting of the wrapping CLI for the decorated function"""
        f.__allow_positional_args__ = True
        return f

    @staticmethod
    def require_options(f):
        f.__allow_positional_args__ = False
        return f

    @staticmethod
    def noncommand(f):
        """mark a function as not to be processed as a command for a CLI (usually a method in a class def)"""
        f.__noncommand__ = True
        return f

    @staticmethod
    def command_name(name):
        """specify a command name string to override the function name. If a prefix is also specified, this will be 
        the last token in the command path for the function after the prefix"""

        def dec(f):
            f.__command_name__ = name
            return f

        return dec

    @staticmethod
    def command_prefix(*prefix):
        """map a function to a nested command. The function name is tokenized, and any tokens from the _tail_ of the
        specified prefix that match those at the _head_ of the function name are only present once in the command path.
        Any remaining (tail) tokens in the function name are also present in the command path.

        The idea is to allow easy nesting of functions under shared command prefixes while yielding succint,
        interpretable command names.

        Example:
            >>> @cli_spec.command_prefix("do", "something")
            >>> def do_something_fast(*args):
            >>>     ...

            This will register `do_something_fast` under the command 'do something fast'

            >>> @cli_spec.command_prefix("do")
            >>> def do_something_fast(*args):
            >>>     ...

            This will register `do_something_fast` under the command 'do something-fast'
        """

        def dec(f):
            f.__command_prefix__ = prefix
            return f

        return dec

    @staticmethod
    def config_subsection(*section):
        def dec(f):
            section_ = _maybe_bool(section, fallback=tuple)
            f.__config_subsections__ = (
                section_ if isinstance(section_, bool) else [section_]
            )
            return f

        return dec

    @staticmethod
    def config_subsections(*sections):
        def dec(f):
            f.__config_subsections__ = [
                tuple(name) if isinstance(name, (list, tuple)) else (name,)
                for name in sections
            ]
            return f

        return dec

    @staticmethod
    def config_top_level(func):
        # get all args straight from the top level of the config hierarchy
        return cli_spec.config_subsection()(func)

    @staticmethod
    def ignore_in_config(*names):
        """mark argument names to be ignored in configuration files for the decorated function"""
        if len(names) == 1 and callable(names[0]):
            # bare decorator
            f = names[0]
            return cli_spec.config_subsection(False)(f)

        def dec(f):
            f.__ignore_in_config__ = _maybe_bool(names)
            return f

        return dec

    @staticmethod
    def parse_config_as_cli(*names):
        """mark argument names to be parsed from configuration files using the parsers applied on the command line"""

        def dec(f):
            f.__parse_config_as_cli__ = _maybe_bool(names)
            return f

        return dec

    @staticmethod
    def ignore_on_cmd_line(*names):
        """mark argument names to be ignored on the command line for the decorated function"""

        def dec(f):
            f.__ignore_on_cmd_line__ = _maybe_bool(names)
            return f

        return dec

    @staticmethod
    def parse_env(**argname_to_envname: str):
        """mark argument names to be parsed from corresponding environment variable names"""

        def dec(f):
            f.__parse_env__ = argname_to_envname
            return f

        return dec

    @staticmethod
    def parse_order(*names):
        """specify the order in which arguments are parsed, for example to allow failing early when some arguments are
        much more costly to parse than others"""

        def dec(f):
            f.__parse_order__ = _validate_parse_order(*names)
            return f

        return dec

    @staticmethod
    def typecheck(*names):
        """mark argument names to be type-checked for the decorated function, or check all arguments if True is passed"""

        def dec(f):
            f.__typecheck__ = _maybe_bool(names)
            return f

        return dec

    @staticmethod
    def metavars(**renames):
        """mark argument names with metavar names as they will appear in the CLI help string, using
        `@cli_spec.metavars(original_var_name=metavar_name, ...)` syntax"""

        def dec(f):
            f.__metavars__ = renames
            return f

        return dec

    @staticmethod
    def output_handler(func):
        """specify a callable to invoke for handling the return value of the decorated function, for example saving to
        disk. The return value is passed as the first argument and any further args will be supplied as keyword args
        parsed from the CLI or config.."""
        if func is None:
            func = NO_OUTPUT_HANDLER
        elif func is NO_OUTPUT_HANDLER:
            pass
        elif not callable(func):
            raise TypeError("func must be callable; got {}".format(type(func)))

        def dec(f):
            f.__output_handler__ = func
            return f

        return dec

    @staticmethod
    def no_output_handler(func):
        return cli_spec.output_handler(NO_OUTPUT_HANDLER)(func)

    @staticmethod
    def exit_codes(codes):
        def dec(f):
            f.__exit_codes__ = codes
            return f

        return dec

    @staticmethod
    def named_groups(**name_to_argnames: Collection[str]):
        def dec(f):
            f.__named_groups__ = {
                name.replace("_", " "): {argnames}
                if isinstance(argnames, str)
                else set(argnames)
                for name, argnames in name_to_argnames.items()
            }
            return f

        return dec


class cli_attrs:
    """retrieve properties set on functions by cli_spec methods"""

    @staticmethod
    def allow_positional_args(f, default=None):
        return getattr(f, "__allow_positional_args__", default)

    @staticmethod
    def require_options(f, default=None):
        if hasattr(f, "__allow_positional_args__"):
            return not cli_attrs.allow_positional_args(f)
        return default

    @staticmethod
    def noncommand(f):
        return getattr(f, "__noncommand__", False)

    @staticmethod
    def command_name(f, default=None):
        return getattr(f, "__command_name__", default)

    @staticmethod
    def command_prefix(f, default=()):
        return getattr(f, "__command_prefix__", default)

    @staticmethod
    def config_subsections(f, default=None):
        return getattr(f, "__config_subsections__", default)

    @staticmethod
    def config_top_level(f):
        return cli_attrs.config_subsections(f) == [()]

    @staticmethod
    def ignore_in_config(f, default=None):
        return getattr(f, "__ignore_in_config__", default)

    @staticmethod
    def parse_config_as_cli(f, default=None):
        return getattr(f, "__parse_config_as_cli__", default)

    @staticmethod
    def ignore_on_cmd_line(f, default=None):
        return getattr(f, "__ignore_on_cmd_line__", default)

    @staticmethod
    def parse_env(f, default=None):
        return getattr(f, "__parse_env__", default)

    @staticmethod
    def parse_order(f):
        return getattr(f, "__parse_order__", None)

    @staticmethod
    def typecheck(f, default=None):
        return getattr(f, "__typecheck__", default)

    @staticmethod
    def metavars(f, default=None):
        return getattr(f, "__metavars__", default)

    @staticmethod
    def output_handler(f, default=None):
        return getattr(f, "__output_handler__", default)

    @staticmethod
    def exit_codes(f):
        return getattr(f, "__exit_codes__", None)

    @staticmethod
    def named_groups(f, default=None):
        return getattr(f, "__named_groups__", default)



# coding:utf-8
from argparse import ONE_OR_MORE, OPTIONAL, ZERO_OR_MORE
from collections import ChainMap, Counter
from functools import lru_cache, reduce
from inspect import Parameter, Signature
from itertools import chain
import operator
from pathlib import Path
from string import punctuation
from typing import Dict, List, Set, Tuple, Union, Optional, Mapping, Generic

from bourbaki.introspection.docstrings import CallableDocs
from ..logging import ProgressLogger, TimedTaskContext
from ..logging.defaults import PROGRESS, ERROR
from ..typed_io.main import ArgSource
from ..typed_io.cli_nargs_ import cli_nargs

VARIADIC_NARGS = (ONE_OR_MORE, OPTIONAL, ZERO_OR_MORE)

_type = tuple({type, *(type(t) for t in (Mapping, Tuple, Generic))})


class LookupOrderConfigError(ValueError):
    def __init__(self, name_or_names):
        self.args = (
            "all names in lookup_order must be in {}; got {}".format(
                tuple(s.name for s in ArgSource), name_or_names
            ),
        )


class LookupOrderRepeated(LookupOrderConfigError):
    def __init__(self, name_or_names):
        self.args = (
            "names in lookup_order must be unique; got {}".format(name_or_names),
        )


_sentinel = object()


class NamedChainMap(ChainMap):
    def __init__(self, *named_maps):
        self.names, maps = zip(*named_maps)
        super().__init__(*maps)

    def get_with_name(self, key, default=_sentinel):
        for name, map_ in zip(self.names, self.maps):
            try:
                return name, map_[key]
            except KeyError:
                continue
        if default is _sentinel:
            raise KeyError(repr(key))

        return None, default


class OutputHandlerSignatureError(TypeError):
    pass


def identity(x):
    return x


def strip_command_prefix(prefix: Union[str, Tuple[str, ...]], funcname: str) -> str:
    if isinstance(prefix, str):
        prefix = (prefix,)

    suffix = tuple(p.replace("-", "_") for p in prefix)
    while not funcname.startswith("_".join(suffix)):
        suffix = suffix[1:]

    strip_prefix = "_".join(suffix)
    funcname = funcname[len(strip_prefix) :].lstrip("_")

    return funcname


def sibling_files(path):
    # for specifying the helper_files arg to the CLI: helper_files=sibling_files(__file__)
    return list(Path(path).parent.glob("*.py"))


def get_task(logger, name, log_level=PROGRESS, error_level=ERROR, time_units="s"):
    if isinstance(logger, ProgressLogger):
        return logger.task(
            name, time_units=time_units, level=log_level, error_level=error_level
        )
    return TimedTaskContext(
        name,
        logger_or_print_func=logger,
        time_units=time_units,
        level=log_level,
        error_level=error_level,
    )


def get_in(subsection, conf, default=_sentinel):
    try:
        return reduce(operator.getitem, subsection, conf)
    except (KeyError, IndexError, TypeError) as e:
        if default is _sentinel:
            raise e
        return default


def update_in(conf, subsection, subconf):
    if not isinstance(subsection, (list, tuple)):
        subsection = (subsection,)

    subconf_ = conf
    for key in subsection:
        subconf__ = subconf_.get(key)
        if subconf__ is None:
            subconf__ = {}
            subconf_[key] = subconf__
        subconf_ = subconf__
    subconf_.update(subconf)


def _help_kwargs_from_docs(
    docs: CallableDocs, long_desc_as_epilog: bool = False, help_: bool = True
):
    if help_:
        kw = {"help": docs.short_desc}
    else:
        kw = {}

    if long_desc_as_epilog:
        kw["description"] = docs.short_desc
        kw["epilog"] = docs.long_desc
    else:
        short = docs.short_desc
        if short:
            if short[-1] in punctuation:
                join = " "
            else:
                join = "; "
            description = join.join(d for d in (short, docs.long_desc) if d)
        else:
            description = docs.long_desc
        kw["description"] = description

    return kw


@lru_cache(None)
def _validate_lookup_order(*lookup_order: ArgSource, include_defaults=True):
    order = []
    missing = []
    for s in lookup_order:
        if isinstance(s, str):
            source = getattr(ArgSource, s, None)
        else:
            source = s

        if not isinstance(source, ArgSource):
            missing.append(s)
        else:
            order.append(source)

    if missing:
        raise LookupOrderConfigError(lookup_order)
    if len(set(lookup_order)) != len(lookup_order):
        raise LookupOrderRepeated(lookup_order)

    if include_defaults and ArgSource.DEFAULTS not in order:
        order.append(ArgSource.DEFAULTS)

    return order


def _validate_parse_order(*parse_order):
    if not parse_order:
        raise ValueError(
            "can't make sense of an empty parse_order: {}".format(parse_order)
        )
    if parse_order.count(Ellipsis) > 1:
        raise ValueError(
            "at most one `...` may be present in parse_order; got {}".format(
                parse_order
            )
        )
    if not all((isinstance(n, str) or n is Ellipsis) for n in parse_order):
        raise TypeError(
            "all elements of parse_order must be str or `...`; got {}".format(
                parse_order
            )
        )
    return parse_order


def _to_name_set(maybe_names, default_set=None, metavar=None):
    if isinstance(maybe_names, str):
        names = {maybe_names}
    elif maybe_names is True:
        # no need to check
        return default_set
    elif not maybe_names:  # None, False
        names = set()
    else:  # collection
        names = set(maybe_names)

    if default_set is not None:
        extra = names.difference(default_set)
        if extra:
            raise NameError(
                "{} must all be in {}; values {} are not".format(
                    metavar or "names", default_set, extra
                )
            )
    return names


def _maybe_bool(names, fallback=_to_name_set):
    if len(names) == 1 and isinstance(names[0], bool):
        return names[0]
    return fallback(names)


def _combined_cli_sig(
    signature: Signature,
    *other_signatures: Signature,
    parse: Optional[Set[str]] = None,
    have_fallback: Optional[Set[str]] = None
):
    all_signatures = (signature, *other_signatures)

    all_names = chain.from_iterable(s.parameters for s in all_signatures)
    if parse is not None:
        all_names = filter(parse.__contains__, all_names)
    name_counts = Counter(all_names)
    common_names = tuple(n for n, c in name_counts.items() if c > 1)
    if common_names:
        raise OutputHandlerSignatureError(
            "signatures share overlapping arg names: {}".format(common_names)
        )

    params = {
        k: []
        for k in (
            Parameter.POSITIONAL_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.VAR_POSITIONAL,
            Parameter.KEYWORD_ONLY,
            Parameter.VAR_KEYWORD,
        )
    }
    for sig in all_signatures:
        for param in sig.parameters.values():
            if parse is None or param.name in parse:
                params[param.kind].append(param)

    # algorithm:
    # sort variadics to the end of the positionals
    # if there are any positional variadics, move all but the first to kw_only, and move any *args to kw_only
    # if there are no positional variadics, move the first *args to the end of positionals, the rest to kw_only
    # if there are more than one **kwargs, move the tail to kw_only

    all_pos = (
        params[Parameter.POSITIONAL_ONLY] + params[Parameter.POSITIONAL_OR_KEYWORD]
    )
    non_variadic_pos, variadic_pos = _separate_variadic_params(all_pos)
    maybe_variadic_pos = []
    if variadic_pos:
        first_variadic = variadic_pos[0]
        last_pos_kind = (
            max(non_variadic_pos[-1].kind, first_variadic.kind)
            if non_variadic_pos
            else first_variadic.kind
        )
        maybe_variadic_pos.append(first_variadic.replace(kind=last_pos_kind))
        params[Parameter.KEYWORD_ONLY].extend(
            p.replace(kind=Parameter.KEYWORD_ONLY) for p in variadic_pos[1:]
        )
    elif params[Parameter.VAR_POSITIONAL]:
        varargs = params[Parameter.VAR_POSITIONAL]
        first_varargs = varargs[0]
        maybe_variadic_pos.append(
            first_varargs.replace(
                kind=Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Tuple[first_varargs.annotation, ...],
            )
        )
        params[Parameter.VAR_POSITIONAL] = varargs[1:]

    for kind in Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD:
        ps = params[kind]
        params[Parameter.KEYWORD_ONLY].extend(
            p.replace(
                kind=Parameter.KEYWORD_ONLY,
                annotation=Tuple[p.annotation, ...]
                if kind is Parameter.VAR_POSITIONAL
                else Dict[str, p.annotation],
            )
            for p in ps
        )

    nondefault_pos, default_pos = _separate_default_params(
        non_variadic_pos, have_fallback=have_fallback
    )
    all_params = (
        nondefault_pos
        + default_pos
        + maybe_variadic_pos
        + params[Parameter.KEYWORD_ONLY]
    )

    # Don't validate because we may have defaults preceding non-defaults at this point
    return Signature(
        all_params,
        return_annotation=signature.return_annotation,
        __validate_parameters__=False,
    )


def _separate_default_params(
    params: List[Parameter], have_fallback: Optional[Set[str]] = None
) -> Tuple[List[Parameter], List[Parameter]]:
    nondefaults, defaults = [], []
    for param in params:
        if param.default is Parameter.empty and (
            not have_fallback or param.name not in have_fallback
        ):
            nondefaults.append(param)
        else:
            defaults.append(param)
    return nondefaults, defaults


def _separate_variadic_params(
    params: List[Parameter]
) -> Tuple[List[Parameter], List[Parameter]]:
    args, variadics = [], []
    for param in params:
        if param.annotation is Parameter.empty:
            args.append(param)
        elif cli_nargs(param.annotation) in VARIADIC_NARGS:
            variadics.append(param)
        else:
            args.append(param)
    return args, variadics



# coding:utf-8
import os
from pathlib import Path
from io import StringIO, BytesIO, IOBase
from typing import Union

FileTypes = (IOBase,)
FileType = Union[FileTypes]

DEFAULT_FILENAME_DATE_FMT = (
    "%Y-%m-%d_%H:%M:%S"
)  # Format for dates appended to files or dirs.
# This will lexsort in temporal order.
DEFAULT_FILENAME_N_DIGITS = 6


def ensure_dir(dir_):
    try:
        if not os.path.exists(dir_):
            os.mkdir(dir_)
        return dir_
    except FileExistsError:
        if os.path.isfile(dir_):
            raise NotADirectoryError("{} exists but is not a directory".format(dir_))
        return None


def get_file(file_or_path, mode=None, allow_dir=False):
    """if a file object is passed, return it unaltered, with a flag indicating that the file should not be closed
    by the caller (the opener may have other uses for it). In this case, if mode is also passed, it is checked against
    the existing file's mode and a ValueError is thrown if they disagree.

    If a string is passed, it is treated as a path and a file at that location is opened and returned with a flag
    indicating that the file should be closed by the caller.
    """
    if not isinstance(file_or_path, FileTypes):
        if isinstance(file_or_path, (str, Path)) and os.path.isdir(file_or_path):
            if not allow_dir:
                raise IsADirectoryError(
                    "allow_dir=False but {} is a directory".format(file_or_path)
                )
            else:
                close = False
                file = Path(file_or_path)
        else:
            close = True
            if mode:
                file = open(file_or_path, mode)
            else:
                file = open(file_or_path)
    else:
        close = False
        file = file_or_path

        if mode is not None:
            if hasattr(file, "mode") and mode != file.mode:
                raise ValueError(
                    "mode {} was requested, but the given file has mode {}".format(
                        mode, file.mode
                    )
                )
            elif isinstance(file, StringIO) and "b" in mode:
                raise ValueError(
                    "mode {} was requested, but the given file is a {}, which supports only text IO".format(
                        mode, type(file)
                    )
                )
            elif isinstance(file, BytesIO) and "b" not in mode:
                raise ValueError(
                    "mode {} was requested, but the given file is a {}, which supports only binary IO".format(
                        mode, type(file)
                    )
                )

    return file, close


def is_newer(file1, file2):
    if file2 is None:
        return True
    mtime1 = os.stat(file1).st_mtime
    if isinstance(file2, (str, bytes, Path)):
        mtime2 = os.stat(file2).st_mtime
    elif isinstance(file2, (float, int)):
        mtime2 = file2
    else:
        raise TypeError(
            "file2 must be str, pathlib.Path, None, float, or int; got {}".format(
                type(file2)
            )
        )
    return mtime1 > mtime2


def dir_prefix_and_ext(prefix, ext=None):
    dir_ = os.path.dirname(prefix)
    if ext is None:
        prefix, ext = os.path.splitext(os.path.basename(prefix))
        if prefix.startswith(".") and ext == "":
            prefix, ext = ext, prefix
    else:
        prefix = os.path.basename(prefix)

    return dir_, prefix, ext


def path_with_ext(file_path, ext=None, disambiguate=False):
    file_path, ext_ = _path_with_ext(file_path, ext)
    if not ext_ and not ext:
        if disambiguate:
            file_path, ext_ = _path_with_ext(disambiguate_path(file_path), ext)
        else:
            raise ValueError(
                "no extension specified for file path {}; try passing one manually via the "
                "`ext` arg or specify `disambiguate=True`".format(file_path)
            )
    else:
        ext_ = ext or ext_

    return file_path, ext_


def _path_with_ext(path, ext=None):
    name, ext_ = os.path.splitext(path)
    if ext_:
        if ext is not None and ext_ != ext:
            raise ValueError(
                "ambiguous extension; config_file has extension {} while ext is {}".format(
                    ext_, ext
                )
            )

    ext_ = ext or ext_
    p, e = name + ext_, ext_
    return p, e


def disambiguate_path(file_path):
    """Find the unique file with path `file_path`, excluding extensions. If there is no such file, raise
    FileNotFoundError"""
    dir_, name, ext = dir_prefix_and_ext(file_path)
    dir_ = dir_ or None  # don't allow empty string for dir
    paths = [path for path in os.listdir(dir_) if os.path.splitext(path)[0] == name]

    if len(paths) == 0:
        raise FileNotFoundError(
            "No file with any extension found at {}".format(file_path)
        )
    elif len(paths) != 1:
        raise FileNotFoundError(
            "Amiguous config path {}; multiple matches found: {}".format(
                file_path, paths
            )
        )
    p = os.path.join(dir_ or "", paths[0])
    return p



# coding:utf-8
import re
from itertools import cycle

datetime_fields = dict(
    Y=r"[0-9]{4}",
    y=r"[0-9]{2}",
    m=r"(?:0[0-9]|1[0-2])",
    b=r"(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)",
    B=r"(?:January|February|March|April|May|June|July|August|September|October|November|December)",
    d=r"(?:[0-2][0-9]|3[01])",
    w=r"[0-6]",
    a=r"(?:Sun?|Mon?|Tue?|Wed?|Thu?|Fri?|Sat?)",
    A=r"(?:Sunday|Monday|Tuesday|Wednesday|Thursday|Friday|Saturday)",
    H=r"(?:[01][0-9]|2[0-4])",
    I=r"(?:0[0-9]|1[0-2])",
    p=r"(?:AM|PM|am|pm)",
    M=r"[0-5][0-9]",
    S=r"[0-5][0-9]",
    f=r"[0-9]{6}",
    z=r"(?:|[-+](?:0[0-9]|1[0-2])(?:[0-5][0-9]))",
    Z=r"(?:|[A-Z]{3})",
    j=r"(?:[0-2][0-9]{2}|3[0-5][0-9]|36[0-6])",
    U=r"(?:[0-4][0-9]|5[0-3])",
    W=r"(?:[0-4][0-9]|5[0-3])",
    X=r"(?:(?:[01][0-9]|2[0-4]):[0-5][0-9]:[0-5][0-9])",
)

date_fmt_field_pat = r"(?!<\\)%({})".format("|".join(datetime_fields))

int_re = r"[0-9]+"
float_re = r"-?[0-9]+\.[0-9]+([eE][+-]?[0-9]+)?"
py_name_re = r"[_a-zA-Z][_a-zA-Z0-9]*"
py_name_dynamic_re = (
    r"<?[_a-zA-Z][_a-zA-Z0-9]*>?"
)  # for objects generated dynamically with no name attribute
py_dot_name_re = r"{name}(?:\.{name})*".format(name=py_name_re)

# note: These are only 'reasonable' patterns for file/path and logger/process names;
# POSIX-compliant names can be much wilder. I keep them simple to allow for a loaders
# parse with little backtracking (as you would get with a lot of '[<big char set>]*'
# subpatterns in your regexes).
# Just be aware that these will break if you give your source files, loggers, and
# processes crazy names or names with whitespace
filename_re = r"[-_.~<>a-zA-Z0-9]+"
pathname_re = r"\/?(?:{name})(?:\/{name})*\/?".format(name=filename_re)
any_name_re = r"[-_.a-zA-Z0-9]*"
upper_name_re = r"[A-Z]+"


def alternating(patterns):
    return "|".join(patterns)


def re_escaped(raw_str):
    return re.sub(r"([(){}\[\].\\?+*])", r"\\\1", raw_str)


def named_group(pattern: str, name: str):
    return r"(?P<{}>{})".format(name, pattern)


def optional(pattern: str):
    return noncapturing_group(pattern) + "?"


def noncapturing_group(pattern: str):
    return r"(?:{})".format(pattern)


def extract_group(match):
    return match.group()


def extract_groups(match: re.match):
    return match.groups()


def with_any_whitespace(pattern, escape=True):
    pattern_ = re_escaped(pattern) if escape else pattern
    return re.sub(r"\s+", r"\s+", pattern_)


def as_word(pattern):
    return "\\b{}\\b".format(noncapturing_group(pattern))


def find_lambda(pattern, s, match_lambda):
    pattern = re.compile(pattern)
    for match in pattern.finditer(s):
        yield match_lambda(match)


def sub_lambda(pattern, s, match_lambda=None, hole_lambda=None):
    pattern = re.compile(pattern)

    def sub_lambda_iter(pattern, s):
        lo, hi = 0, 0
        for match in re.finditer(pattern, s):
            hi, _ = match.span()
            yield s[lo:hi]
            yield match
            lo = _
        yield s[lo:]

    if hole_lambda is None:
        hole_lambda = lambda x: x
    if match_lambda is None:
        match_lambda = extract_group

    holematches = sub_lambda_iter(pattern, s)
    holematches = list(holematches)
    funcs = cycle((hole_lambda, match_lambda))

    return "".join(f(x) for f, x in zip(funcs, holematches))


def datetime_regex(date_fmt: str):
    def get_field_re(match):
        char = match.group(1)
        pat = datetime_fields.get(char)
        if not pat:
            raise ValueError("Unrecognized datetime format char: '%s'" % char)
        return noncapturing_group(pat)

    return sub_lambda(date_fmt_field_pat, date_fmt, get_field_re, re_escaped)



# coding:utf-8
from typing import Callable, Union, Optional as Opt
from time import time
from datetime import datetime, timedelta
from functools import partial
from logging import Logger, getLogger
from .defaults import PROGRESS, ERROR, DEFAULT_LOG_DATE_FMT
from .helpers import validate_log_level_int

unit_names = dict(
    s="s",
    seconds="s",
    ms="ms",
    milliseconds="ms",
    mus="μs",
    us="μs",
    μs="μs",
    microseconds="μs",
    ns="ns",
    nanoseconds="ns",
    m="m",
    mins="m",
    minutes="m",
    h="h",
    hrs="h",
    hours="h",
    days="d",
    d="d",
)

timedelta_units_kw = dict(
    s="seconds", ms="milliseconds", μs="microseconds", m="minutes", h="hours", d="days"
)

from_seconds_multipliers = dict(
    s=1.0, ms=1e3, μs=1e6, ns=1e9, m=1.0 / 6e1, h=1.0 / 36e2, d=1.0 / 864e2
)


class TimedTaskContext:
    def __init__(
        self,
        job_name: str,
        total_tasks: Opt[int] = None,
        task_units: Opt[str] = None,
        logger_or_print_func: Opt[Union[Logger, Callable[[str], type(None)]]] = print,
        level=PROGRESS,
        error_level=ERROR,
        time_units: str = "s",
        date_fmt=DEFAULT_LOG_DATE_FMT,
    ):
        try:
            self.time_units = unit_names[time_units]
            # timedelta won't take nanoseconds on some systems; we multiply microseconds by 1000
            if time_units == "ns":
                self.timedelta_kw = "microseconds"
                self.timedelta_multiplier = 1000.0
            else:
                self.timedelta_kw = timedelta_units_kw[time_units]
                self.timedelta_multiplier = 1.0
            self.multiplier = from_seconds_multipliers[self.time_units]
        except KeyError:
            raise ValueError(
                "{} is not a valid time units identifer; choose one of {}".format(
                    time_units, set(unit_names)
                )
            )

        if isinstance(logger_or_print_func, Logger) or logger_or_print_func is None:
            level = validate_log_level_int(level)
            error_level = validate_log_level_int(error_level)
            if logger_or_print_func is None:
                logger = getLogger(job_name)
            else:
                logger = logger_or_print_func
            self.logger = logger
            self.print = partial(logger.log, level)
            self.error = partial(logger.log, error_level)
        else:
            assert callable(
                logger_or_print_func
            ), "print_func must be callable or a name for a logger"
            self.logger = None
            self.print = logger_or_print_func
            self.error = self.print

        self.total = int(total_tasks) if total_tasks is not None else None
        self.completed = 0
        self.task_units = str(task_units) if task_units is not None else ""

        self.date_fmt = date_fmt

        self.job_name = str(job_name)
        self.start = None

        self.n_digits = 4

    def elapsed_since(self, start):
        end = time()
        elapsed = self.multiplier * (end - start)
        return elapsed

    def total_elapsed(self):
        return self.elapsed_since(self.start)

    def report_elapsed(self):
        self.print(self._elapsed_time_message())

    def report_progress(self, n_tasks: int, timing_info=True):
        self.completed += n_tasks
        msg = self._progress_message()
        if timing_info:
            msg = "{}; {}".format(msg, self._timing_info_message(prefix=False))

        self.print(msg)

    @property
    def is_finished(self):
        if self.total is None:
            return False
        return self.completed >= self.total

    def _elapsed_time_message(self, start=None, prefix=True):
        if start is None:
            start = self.start
        elapsed = self.elapsed_since(start)
        return "{}elapsed time {}{}".format(
            "{}{}: ".format(
                self.job_name, " " + self.task_units if self.task_units else ""
            )
            if prefix
            else "",
            round(elapsed, self.n_digits),
            self.time_units,
        )

    def _progress_message(self):
        if self.total:
            return "{}: completed {} of {}".format(
                self.job_name, self.completed, self.total
            )
        else:
            return "{}: completed {}".format(self.job_name, self.completed)

    def _timing_info_message(self, prefix=True, estimate=True, completed=None):
        if completed is None:
            completed = self.completed
        rate = completed / self.total_elapsed()
        if estimate and self.total:
            remaining = self.total - self.completed
            expected_end = datetime.now() + timedelta(
                **{self.timedelta_kw: remaining * rate / self.timedelta_multiplier}
            )

            msg = "average rate {} {}/{}; expected completion {}".format(
                round(rate, self.n_digits),
                self.task_units or "tasks",
                self.time_units,
                expected_end.strftime(self.date_fmt),
            )
        else:
            msg = "average rate {} {}/{}".format(
                round(rate, self.n_digits), self.task_units or "tasks", self.time_units
            )

        if prefix:
            msg = "{}: {}".format(self.job_name, msg)

        return msg

    def _error_message(self):
        if self.total is not None:
            return "FAILED {}; completed {} {}of {}".format(
                self.job_name,
                self.completed,
                self.task_units + " " if self.task_units else "",
                self.total,
            )
        return "FAILED {}".format(self.job_name)

    def _reset_time(self):
        self.start = time()

    def __enter__(self):
        if self.total is None:
            self.print(
                "STARTED {} at {}".format(
                    self.job_name, datetime.now().strftime(self.date_fmt)
                )
            )
        else:
            self.print(
                "STARTED {} {} {}at {}".format(
                    self.job_name,
                    self.total,
                    self.task_units + " " if self.task_units else "",
                    datetime.now().strftime(self.date_fmt),
                )
            )
        self._reset_time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None or exc_val is not None:
            msg = self._error_message()
            self.error(msg + "; {}".format(exc_val or exc_type))
            raise (exc_val or exc_type)
        else:
            if self.total is not None:
                self.print(
                    "FINISHED {}; {}".format(
                        self._elapsed_time_message(),
                        self._timing_info_message(
                            prefix=False,
                            estimate=False,
                            completed=self.total or self.completed,
                        ),
                    )
                )
            else:
                self.print("FINISHED {}".format(self._elapsed_time_message()))


timed_context = TimedTaskContext



# coding:utf-8
from logging import _nameToLevel, _levelToName


def validate_log_level(level):
    assert (
        level in _nameToLevel or level in _levelToName or level == "NOTSET"
    ), "'{}' is not a valid logging level".format(level)
    return level


def validate_log_level_int(level):
    level = validate_log_level(level)
    return level if level in _levelToName else _nameToLevel.get(level, level)


def validate_log_level_str(level):
    level = validate_log_level(level)
    return level if level in _nameToLevel else _levelToName.get(level, level)



# coding:utf-8
from typing import Callable
from collections.abc import Mapping
from collections import OrderedDict
from functools import update_wrapper
from inspect import signature, Parameter
from itertools import chain
from logging import getLogger
from cytoolz import valmap
from bourbaki.introspection.imports import import_object, import_type
from bourbaki.introspection.types import (
    typetypes,
    eval_forward_refs,
    concretize_typevars,
    get_generic_origin,
    get_param_dict,
    reparameterized_bases,
    fully_concretize_type,
    issubclass_generic,
)
from bourbaki.introspection.types.compat import NEW_TYPING, _GenericAlias
from bourbaki.introspection.typechecking import isinstance_generic
from bourbaki.introspection.callables import to_bound_method_signature, get_globals
from bourbaki.introspection.classes import most_specific_constructor
from bourbaki.introspection.wrappers import cached_getter, lru_cache_sig_preserving
from .exceptions import ExceptionReprMixin, ConfigTypedInputError

config_decoder = None  # this will be imported later to avoid circular imports
logger = getLogger(__name__)
CLASSPATH_KEY = "__classpath__"
ARGS_KEY = "__args__"
KWARGS_KEY = "__kwargs__"
CONSTRUCTOR_KEY = "__constructor__"
INFLATABLE_CONFIG_KEYS = frozenset(
    [CLASSPATH_KEY, ARGS_KEY, KWARGS_KEY, CONSTRUCTOR_KEY]
)
NON_VARIADIC_KINDS = (
    Parameter.POSITIONAL_OR_KEYWORD,
    Parameter.KEYWORD_ONLY,
    Parameter.POSITIONAL_ONLY,
)
TYPE_TYPES = (type, *typetypes)


class ConfigInflationError(ValueError, ExceptionReprMixin):
    def __init__(self, func, argname, value, exc):
        super().__init__()
        self.func, self.argname, self.value, self.exc = func, argname, value, exc

    def __str__(self):
        return "Decoding config value {} for argument '{}' of function {} raised {}".format(
            repr(self.value), self.argname, self.func, self.exc
        )


def _is_inflatable_config(obj):
    return (
        isinstance(obj, Mapping)
        and (CLASSPATH_KEY in obj or CONSTRUCTOR_KEY in obj)
        and INFLATABLE_CONFIG_KEYS.issuperset(obj.keys())
    )


def inflate_config(conf, target_type=None):
    if _is_inflatable_config(conf):
        try:
            return instance_from(**conf, target_type=target_type)
        except Exception as e:
            raise ConfigTypedInputError(target_type, conf, e)
    return conf


def instance_from(
    __classpath__=None,
    __args__=None,
    __kwargs__=None,
    __constructor__=None,
    target_type=None,
):
    """
    Instantiate general values from configuration files. E.g. with the json-compatible configuration
        conf = {"__classpath__":"sklearn.cluster.KMeans", "__kwargs__": {"n_clusters": 10, "n_init": 10}},
        instance_from(**conf) returns an sklearn.cluster.KMeans clustering model.
    :param __classpath__: str, optional path to the class to be instantiated. If no __constructor__ is passed, this
        will be called directly, otherwise it will be used for a final instance check.
    :param __args__: optional positional args to be passed to the class constructor (or function if classpath refers to
        a general callable)
    :param __kwargs__: optional keyword args to be passed to the class constructor (or function if classpath refers to a
        general callable)
    :param __constructor__: optional path to a callable to call to construct the desired instance
    :param target_type: optional type to check the classpath against before attempting to inflate an instance
    :return: an instance of the class (or the results of calling the function) identified by classpath
    """
    logger.debug(
        "Attempting to instantiate {} instance with{} args {} and kwargs {}".format(
            __classpath__,
            " constructor {},".format(__constructor__)
            if __constructor__ is not None
            else "",
            __args__,
            __kwargs__,
        )
    )

    if __classpath__ is not None:
        cls = import_type(__classpath__)
        if not isinstance(cls, TYPE_TYPES):
            raise TypeError(
                "classpath {} does not specify a class or type; got {}".format(
                    __classpath__, cls
                )
            )
        # don't waste time on the construction if the specified type is incorrect
        if target_type is not None:
            target_type_ = concretize_typevars(target_type)
            if not issubclass_generic(cls, target_type_):
                raise TypeError(
                    "classpath {} does not specify a generic subclass of the target type {}".format(
                        __classpath__, target_type
                    )
                )
        # only check the instance if the constructor is other than the class itself
        instance_check = __constructor__ is not None
    else:
        instance_check = False
        cls = None

    if __constructor__ is not None:
        constructor = import_object(__constructor__)
        if not callable(constructor):
            raise TypeError(
                "constructor {} does not specify a callable; got {}".format(
                    __constructor__, constructor
                )
            )
    elif cls is None:
        raise ValueError("Must pass either __classpath__ or __constructor__")
    else:
        constructor = cls

    wrapper = typed_config_callable(constructor)

    if __kwargs__ is None and __args__ is None:
        obj = wrapper()
    elif __kwargs__ is None:
        obj = wrapper(*__args__)
    elif __args__ is None:
        obj = wrapper(**__kwargs__)
    else:
        obj = wrapper(*__args__, **__kwargs__)

    if instance_check and not isinstance_generic(obj, cls):
        raise TypeError(
            "Inflation using constructor {} resulted in a {} instance; expected {}".format(
                constructor, type(obj), cls
            )
        )

    logger.info(
        "Instantiated {} instance successfully{}".format(
            __classpath__,
            " with constructor {}".format(__constructor__)
            if __constructor__ is not None
            else "",
        )
    )

    return obj


class TypedConfigCallable:
    called = False
    __signature__ = None

    def __init__(self, func: Callable, param_dict=None, ignore_args=()):
        func_for_sig = func

        if isinstance(func, TYPE_TYPES):
            cls_for_sig, func_for_sig = most_specific_constructor(
                func, return_class=True
            )
            skip_first_arg = not isinstance(func_for_sig, staticmethod)

            _cls_for_sig = get_generic_origin(cls_for_sig)
            if param_dict is None:
                for base in chain((func,), reparameterized_bases(func)):
                    if get_generic_origin(base) is _cls_for_sig:
                        param_dict = get_param_dict(base)
        else:
            skip_first_arg = False

        try:
            sig = signature(func_for_sig)
        except TypeError as e:
            raise e
        except ValueError:
            # builtins and C extensions
            sig = None

        if skip_first_arg:
            sig = to_bound_method_signature(sig)

        update_wrapper(self, func_for_sig)
        self.__signature__ = sig
        self.func = func
        self.func_for_sig = func_for_sig
        self.param_dict = param_dict
        self.skip_first_arg = skip_first_arg
        self.ignore_args = set(ignore_args)

    @staticmethod
    def concretize_annotation(ann, globals_):
        return concretize_typevars(eval_forward_refs(ann, globals_))

    @property
    @cached_getter
    def decoders(self):
        # importing in a function call is ugly, but better than a giant source file, which is the only other way
        # of getting around the circular import
        global config_decoder
        if not TypedConfigCallable.called:
            from bourbaki.application.typed_io.config_decode import config_decoder

            TypedConfigCallable.called = True

        if self.__signature__ is None:
            return None

        globals_ = get_globals(self.func_for_sig)
        return OrderedDict(
            (
                name,
                config_decoder(
                    fully_concretize_type(p.annotation, self.param_dict, globals_)
                ),
            )
            for name, p in self.__signature__.parameters.items()
            if name not in self.ignore_args
        )

    def __call__(self, *args, **kwargs):
        if self.__signature__ is not None:
            bound = self.__signature__.bind(*args, **kwargs)
            bound_args = bound.arguments
            params = self.__signature__.parameters
            decoders = self.decoders

            for name, arg in bound_args.items():
                param = params[name]
                try:
                    if param.kind in NON_VARIADIC_KINDS:
                        bound_args[name] = decoders[name](arg)
                    elif param.kind is Parameter.VAR_KEYWORD:
                        bound_args[name] = valmap(decoders[name], arg)
                    elif param.kind is Parameter.VAR_POSITIONAL:
                        bound_args[name] = tuple(map(decoders[name], arg))
                except Exception as e:
                    raise ConfigInflationError(self.func, name, arg, e)

            bound.apply_defaults()
            args, kwargs = bound.args, bound.kwargs

        self.called = True
        return self.func(*args, **kwargs)


@lru_cache_sig_preserving(None)
def typed_config_callable(func, param_dict=None):
    return TypedConfigCallable(func, param_dict)



# coding:utf-8
import typing
import enum
from inspect import signature, Signature, Parameter
from itertools import chain
from typing_inspect import is_optional_type
from bourbaki.introspection.generic_dispatch import (
    GenericTypeLevelSingleDispatch,
    UnknownSignature,
    const,
)
from bourbaki.introspection.generic_dispatch_helpers import UnionWrapper, LazyWrapper
from bourbaki.introspection.classes import (
    classpath,
    parameterized_classpath,
    most_specific_constructor,
)
from bourbaki.introspection.callables import (
    has_varargs,
    get_globals,
    get_callable_params,
)
from bourbaki.introspection.wrappers import lru_cache_sig_preserving
from bourbaki.introspection.types import (
    issubclass_generic,
    is_top_type,
    fully_concretize_type,
    get_generic_params,
    get_named_tuple_arg_types,
    BuiltinAtomic,
    NonStrCollection,
    NonStdLib,
    LazyType,
    NamedTupleABC,
)
from .exceptions import (
    ConfigIOUndefined,
    ConfigIOUndefinedForKeyType,
    ConfigCollectionKeysNotAllowed,
)
from .utils import (
    type_spec,
    default_repr_values,
    byte_repr,
    any_repr,
    ellipsis_,
    repr_type,
    Empty,
    unq,
)
from .inflation import CONSTRUCTOR_KEY, CLASSPATH_KEY, KWARGS_KEY, ARGS_KEY
from .parsers import EnumParser

NoneType = type(None)

null_config_repr = None
null_config_repr_str = "null"

bytes_config_repr = [byte_repr, ellipsis_]
bool_config_repr = type_spec(bool)

bytes_config_key_repr = "b'\\x{}{}'".format(byte_repr, ellipsis_)
bool_config_key_repr = "true|false"

config_repr_values = default_repr_values.copy()
config_repr_values.update(
    [
        (typing.ByteString, bytes_config_repr),
        (bool, bool_config_repr),
        (NoneType, null_config_repr),
    ]
)

config_key_repr_values = default_repr_values.copy()
config_key_repr_values.update(
    [
        (bool, bool_config_key_repr),
        (typing.ByteString, bytes_config_key_repr),
        # don't try to parse complex types for keys
        (typing.Any, any_repr),
        (NoneType, null_config_repr),
    ]
)


ONLY_REQUIRED_ARGS = False
LITERAL_DEFAULTS = True


def config_repr_all_args():
    global ONLY_REQUIRED_ARGS
    ONLY_REQUIRED_ARGS = False


def config_repr_only_required_args():
    global ONLY_REQUIRED_ARGS
    ONLY_REQUIRED_ARGS = True


def config_repr_literal_defaults():
    global LITERAL_DEFAULTS
    LITERAL_DEFAULTS = True


def config_repr_non_literal_defaults():
    global LITERAL_DEFAULTS
    LITERAL_DEFAULTS = False


@lru_cache_sig_preserving(None)
def config_repr_callable(
    f: typing.Callable, *args: type, only_required_args=None, literal_defaults=None
):
    if args:
        params = get_callable_params(f)
        param_dict = dict(zip(params, args))
    else:
        param_dict = None

    repr_, sig = config_repr_callable_args(
        f,
        param_dict,
        skip_self=False,
        only_required_args=only_required_args,
        literal_defaults=literal_defaults,
    )
    repr_[CONSTRUCTOR_KEY] = classpath(f)

    return_ = sig.return_annotation
    if return_ not in (Parameter.empty, object, typing.Any):
        return_ = fully_concretize_type(return_, param_dict, get_globals(f))
        repr_[CLASSPATH_KEY] = parameterized_classpath(return_)

    return repr_


@lru_cache_sig_preserving(None)
def config_repr_class(
    t: typing.Type,
    *args: type,
    only_required_args=None,
    literal_defaults=None,
    constructor: typing.Optional[typing.Union[str, typing.Callable]] = None
):
    skip_self = False
    if constructor is None:
        # don't use just the type for the signature; in cases where Generic is subclassed for instance you'll
        # get an uninformative *args, **kwargs signature
        init = most_specific_constructor(t)
        skip_self = True
        add_constructor = False
    elif isinstance(constructor, str):
        try:
            init = getattr(t, constructor)
        except AttributeError:
            raise AttributeError(
                "constructor name '{}' is not an attribute of type {}".format(
                    constructor, t
                )
            )
        add_constructor = True
    elif callable(constructor):
        init = constructor
        add_constructor = True
    else:
        raise TypeError(
            "if passed, constructor must be a callable or attribute name; got {}".format(
                type(constructor)
            )
        )

    params = get_generic_params(t)
    param_dict = dict(zip(params, args))

    repr_, sig = config_repr_callable_args(
        init,
        param_dict,
        skip_self=skip_self,
        only_required_args=only_required_args,
        literal_defaults=literal_defaults,
    )

    if args:
        if len(args) == 1:
            args = args[0]
        type_ = t[args]
    else:
        type_ = t

    repr_[CLASSPATH_KEY] = parameterized_classpath(
        fully_concretize_type(type_, param_dict, get_globals(t))
    )
    if add_constructor:
        repr_[CONSTRUCTOR_KEY] = classpath(init)
    return repr_


def config_repr_callable_args(
    init,
    param_dict=None,
    skip_self=False,
    only_required_args=None,
    literal_defaults=None,
):
    if only_required_args is None:
        only_required_args = ONLY_REQUIRED_ARGS
    if literal_defaults is None:
        literal_defaults = LITERAL_DEFAULTS

    if isinstance(init, Signature):
        sig = init
        globals_ = None
    else:
        sig = signature(init)
        globals_ = get_globals(init)

    concretize = param_dict or globals_
    append_positionals = has_varargs(sig)
    arg_reprs, kwarg_reprs = [], {}
    parameter_iter = iter(sig.parameters.items())
    if skip_self:
        next(parameter_iter)

    for name, param in parameter_iter:
        if literal_defaults and (param.default not in (Parameter.empty, None)):
            r = param.default
        else:
            if concretize:
                type_ = fully_concretize_type(param.annotation, param_dict, globals_)
            else:
                type_ = param.annotation
            r = config_repr(type_)

        if param.kind is Parameter.POSITIONAL_ONLY:
            arg_reprs.append(r)
        elif param.kind is Parameter.VAR_POSITIONAL:
            arg_reprs.extend([r, ellipsis_])
        elif param.kind is Parameter.VAR_KEYWORD:
            kwarg_reprs[ellipsis_] = r
        elif (
            only_required_args
            and param.default is not Parameter.empty
            and is_optional_type(param.annotation)
        ):
            # only skip once we're past the positionals, to avoid skipping args
            continue
        elif param.kind is Parameter.KEYWORD_ONLY:
            kwarg_reprs[name] = r
        else:
            if append_positionals:
                arg_reprs.append(r)
            else:
                kwarg_reprs[name] = r

    repr_ = {KWARGS_KEY: kwarg_reprs}
    if arg_reprs:
        repr_[ARGS_KEY] = arg_reprs

    return repr_, sig


# the main method

config_repr = GenericTypeLevelSingleDispatch(
    "config_repr", isolated_bases=[typing.Union, NonStdLib]
)

config_key_repr = GenericTypeLevelSingleDispatch(
    "config_key_repr", isolated_bases=[typing.Union, NonStdLib]
)

config_repr.register(BuiltinAtomic)(type_spec)


@config_repr.register_all(typing.Any, typing.Generic, NonStdLib)
def config_repr_any(t, *types, **kw):
    if is_top_type(t):
        return any_repr
    return config_repr_class(t, *types, **kw)


class _ConfigReprUnion(UnionWrapper):
    tolerate_errors = (ConfigIOUndefined, UnknownSignature)
    tolerate_errors_call = ()
    exc_class = ConfigIOUndefined

    @staticmethod
    def getter(t):
        return const(config_repr(t))

    @staticmethod
    def reduce(reprs):
        reprs = list(reprs)
        if all(isinstance(r, (str, NoneType)) for r in reprs):
            return _ConfigKeyReprUnion.reduce(r or null_config_repr_str for r in reprs)
        return list(chain.from_iterable((r, "OR") for r in reprs))[:-1]


ConfigReprUnion = lru_cache_sig_preserving(None)(_ConfigReprUnion)


@config_repr.register(typing.Union)
def config_repr_union(u, *types):
    return ConfigReprUnion(u, *types)(None)


@config_repr.register(typing.Tuple)
def config_repr_tuple(t, *types):
    if types and types[-1] is not Ellipsis:
        return list(map(config_repr, types))
    elif not types:
        return [any_repr, ellipsis_]
    else:
        return [config_repr(types[0]), ellipsis_]


@config_repr.register(NamedTupleABC)
def config_repr_namedtuple(t, *types):
    if not types:
        types = get_named_tuple_arg_types(t)
    return dict(zip(t._fields, map(config_repr, types)))


@config_repr.register(typing.Mapping)
def config_repr_mapping(m, k=object, v=object):
    # have to check this here because the key type param is invariant; we can't resolve from a base class
    if issubclass_generic(k, NonStrCollection):
        raise ConfigCollectionKeysNotAllowed((m, k, v))
    try:
        key_repr = config_key_repr(k)
    except UnknownSignature:
        raise ConfigIOUndefinedForKeyType((m, k, v))
    else:
        return {key_repr: config_repr(v), ellipsis_: ellipsis_}


@config_repr.register(typing.Collection)
def config_repr_seq(s, t=Empty, e=ellipsis_):
    return [config_repr(t), ellipsis_]


@config_repr.register(LazyType)
def lazy_config_repr(lazy, ref):
    return LazyConfigRepr(lazy, ref)()


config_repr.register(typing.Type)(repr_type)
config_key_repr.register(typing.Type)(repr_type)


class LazyConfigRepr(LazyWrapper):
    @staticmethod
    def getter(type_):
        return const(config_repr(type_))


config_repr.register_from_mapping(config_repr_values, as_const=True)


@config_repr.register_all(enum.Enum, enum.IntEnum)
@config_key_repr.register_all(enum.Enum, enum.IntEnum)
def config_repr_enum(enum_):
    return EnumParser(enum_).config_repr()


class _ConfigKeyReprUnion(UnionWrapper):
    tolerate_errors = (ConfigIOUndefined, UnknownSignature)
    tolerate_errors_call = ()
    exc_class = ConfigIOUndefinedForKeyType
    getter = config_key_repr

    @staticmethod
    def reduce(reprs):
        return ' OR '.join(unq(reprs))

    @staticmethod
    def getter(t):
        return const(config_key_repr(t))


ConfigKeyReprUnion = lru_cache_sig_preserving(None)(_ConfigKeyReprUnion)


@config_key_repr.register(typing.Union)
def config_key_repr_union(u, *types):
    return ConfigKeyReprUnion(u, *types)(None)


config_key_repr.register_all(int, float)(type_spec)

config_key_repr.register_from_mapping(config_key_repr_values, as_const=True)



# coding:utf-8
from .main import CommandLineInterface
from .helpers import sibling_files
from bourbaki.application.typed_io import ArgSource, CLI, CONFIG, ENV
from .decorators import cli_spec
from .actions import InstallShellCompletionAction, InfoAction, PackageVersionAction
from bourbaki.application.typed_io.utils import File, TextFile, BinaryFile
from bourbaki.introspection.imports import lazy_imports, from_, import_



# coding:utf-8
import os
import multiprocessing
from itertools import zip_longest
from logging import Logger


def get_nproc(n):
    if not n:
        return None
    elif not isinstance(n, int):
        raise TypeError("if truthy, n must be an integer")

    ncores = os.cpu_count()
    if n > 0:
        nproc = min(n, ncores)
    elif n < 0:
        nproc = max(1, ncores + n + 1)
    elif n == 0:
        raise ValueError("nproc must be positive or negative; got 0")

    return nproc


def init_logger(logger_: Logger):
    proc = multiprocessing.current_process()
    global logger
    logger = logger_
    logger.info("process {} initialized successfully".format(proc))


def get_pool(nproc, logger, init=None):
    """Put a logger in the global namespaces of your processes at init.
    Make sure your logger's handler is process-safe, e.g. application.logging.MultiProcStreamHandler or
    application.logging.MultiProcRotatingFileHandler"""
    if not isinstance(logger, Logger):
        raise TypeError(
            "You must pass a logging.Logger instance for logger; got {}".format(
                type(logger)
            )
        )

    if init is None:
        init = init_logger
        initargs = (logger,)
    else:
        if not callable(init):
            raise TypeError(
                "init must be a callable to be executed at the initialization of a process"
            )
        init = apply_all(init_logger, init)
        initargs = [[logger]]

    return multiprocessing.Pool(nproc, initializer=init, initargs=initargs)


class apply_all:
    """Acts as a wrapper function to apply all of a sequence of functions a the initialization of a process worker."""

    def __init__(self, *funcs):
        self.fs = tuple(funcs)

    def __call__(self, *args):
        """pass one (args, kwargs) tuple or kwargs dict for each function to act on. If less args are supplied than
        there are functions, the tail functions are called without arguments. If more tuples are supplied than there
        are functions, an error is thrown"""

        results = []
        for f, a in zip_longest(self.fs, args, fillvalue=((), {})):
            if isinstance(a, tuple):
                if len(a) == 2:
                    args, kwargs = a
                    if args is None:
                        args = ()
                    if kwargs is None:
                        kwargs = {}
                else:
                    raise ValueError(
                        "if args are tuples, they must be 2-tuples of (args_tuple, kwargs_dict)"
                    )
            if isinstance(a, dict):
                args, kwargs = (), a
            elif isinstance(a, list):
                args, kwargs = a, {}
            elif a is None:
                args, kwargs = (), {}

            results.append(f(*args, **kwargs))

        return results



# coding:utf-8
import typing
import enum
from itertools import repeat
from bourbaki.introspection.generic_dispatch import GenericTypeLevelSingleDispatch
from bourbaki.introspection.types import (
    issubclass_generic,
    is_named_tuple_class,
    get_named_tuple_arg_types,
    is_top_type,
    LazyType,
    NonStrCollection,
)
from .cli_nargs_ import cli_nargs
from .exceptions import CLIIOUndefined, CLINestedCollectionsNotAllowed
from .parsers import bool_constants, EnumParser
from .utils import (
    byte_repr,
    any_repr,
    classpath_function_repr,
    default_repr_values,
    repr_type,
)
from .utils import type_spec, KEY_VAL_JOIN_CHAR, to_str_cli_repr

NoneType = type(None)

bool_cli_repr = "{{{}}}".format("|".join(bool_constants))

cli_repr_values = default_repr_values.copy()
cli_repr_values.update(
    [
        (typing.ByteString, byte_repr),
        (bool, bool_cli_repr),
        (typing.Callable, classpath_function_repr),
    ]
)


cli_repr = GenericTypeLevelSingleDispatch(__name__, isolated_bases=[typing.Union])


# base generic handlers


@cli_repr.register(typing.Any)
def default_cli_repr(type_, *args):
    return any_repr if is_top_type(type_) else type_spec(type_)


@cli_repr.register(typing.Union)
def cli_repr_union(u, *types):
    def inner(types_):
        cache = set()
        for t in types_:
            if t is NoneType or t is None:
                continue
            try:
                repr_ = cli_repr(t)
            except CLIIOUndefined:
                continue
            else:
                if repr_ not in cache:
                    cache.add(repr_)
                    yield repr_

    reprs = list(inner(types))
    if not reprs:
        raise CLIIOUndefined((u, *types))
    if all(isinstance(r, str) for r in reprs):
        return "|".join(reprs)
    else:
        lens = set(len(t) for t in reprs if not isinstance(t, str))
        if len(lens) > 1:
            raise ValueError(
                "can't format CLI representation for {}; got sub-reprs {}. for a Union, the reprs "
                "of the types must all be str or tuples of a single length".format(
                    u[types], reprs
                )
            )
        maxlen = max(lens) if lens else 1
        tups = zip(*(repeat(s, maxlen) if isinstance(s, str) else s for s in reprs))
        return tuple("|".join(tup) for tup in tups)


@cli_repr.register(typing.Tuple)
def cli_repr_tuple(t, *types):
    if not types and is_named_tuple_class(t):
        types = get_named_tuple_arg_types(t)

    if types and Ellipsis not in types:
        return tuple(to_str_cli_repr(cli_repr(t), cli_nargs(t)) for t in types)
    elif not types:
        t = typing.Any
    else:
        t = types[0]

    return cli_repr(t)


@cli_repr.register(typing.Mapping)
def cli_repr_mapping(m, k, v):
    if issubclass_generic(k, NonStrCollection) or issubclass_generic(
        v, NonStrCollection
    ):
        raise CLINestedCollectionsNotAllowed((m, k, v))
    return "{k}{j}{v}".format(k=cli_repr(k), v=cli_repr(v), j=KEY_VAL_JOIN_CHAR)


# uni-typed tuples are variable-length, and sets can be parsed from sequences as well
@cli_repr.register(typing.Collection)
def cli_repr_seq(s, t=typing.Any):
    if issubclass_generic(t, NonStrCollection):
        raise CLINestedCollectionsNotAllowed((s, t))
    return cli_repr(t)


@cli_repr.register(typing.Collection[NonStrCollection])
@cli_repr.register(typing.Tuple[NonStrCollection, ...])
def cli_repr_nested_seq(s, t=typing.Any, ellipsis=Ellipsis):
    # this allows going down one level of nesting in the case of options where an append mode can parse
    return cli_repr(t)


@cli_repr.register_all(enum.Enum, enum.IntEnum)
def cli_repr_enum(enum_):
    return EnumParser(enum_).cli_repr()


cli_repr.register(typing.Type)(repr_type)


@cli_repr.register(LazyType)
def cli_repr_lazy(l, t):
    return type_spec(getattr(t, "__forward_arg__", t))


# concrete handlers

cli_repr.register_from_mapping(cli_repr_values, as_const=True)



from typing import (
    Callable,
    Collection,
    Dict,
    Set,
    List,
    Mapping,
    NamedTuple,
    Pattern,
    Iterable,
    Optional,
    Union,
    Any,
)
from collections import ChainMap
from functools import singledispatch
from enum import Enum
from itertools import chain
from inspect import Parameter, Signature
import re

from bourbaki.introspection.callables import (
    leading_positionals,
    most_specific_constructor,
)
from .decorators import cli_attrs
from .helpers import _validate_parse_order, _type


class ArgKind(Enum):
    PositionalOnly = Parameter.POSITIONAL_ONLY
    PositionalOrKeyword = Parameter.POSITIONAL_OR_KEYWORD
    KeywordOnly = Parameter.KEYWORD_ONLY
    VarArgs = Parameter.VAR_POSITIONAL
    VarKwargs = Parameter.VAR_KEYWORD
    AllPositional = (
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.POSITIONAL_ONLY,
        Parameter.VAR_POSITIONAL,
    )
    StrictPositional = (Parameter.POSITIONAL_ONLY, Parameter.VAR_POSITIONAL)
    AllKeyWord = (
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.KEYWORD_ONLY,
        Parameter.VAR_KEYWORD,
    )
    StrictKeyWord = (Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD)


_ParameterKind = type(Parameter.POSITIONAL_OR_KEYWORD)

ArgNameSpec = Union[str, Pattern[str], ArgKind, _ParameterKind]

AnyArgNameSpec = Union[bool, ArgNameSpec, Collection[ArgNameSpec]]


class CLISignatureSpec(NamedTuple):
    ignore_on_cmd_line: Optional[AnyArgNameSpec] = None
    ignore_in_config: Optional[AnyArgNameSpec] = None
    parse_config_as_cli: Optional[AnyArgNameSpec] = None
    typecheck: Optional[AnyArgNameSpec] = None
    parse_env: Optional[Dict[str, str]] = None
    metavars: Optional[Dict[str, str]] = None
    named_groups: Optional[Dict[str, Collection[str]]] = None
    parse_order: Optional[List[Union[str, type(Ellipsis)]]] = None
    require_options: Optional[bool] = None

    @classmethod
    def from_callable(cls, func: Callable) -> "CLISignatureSpec":
        if isinstance(func, _type):
            constructor = most_specific_constructor(func)
        else:
            constructor = None

        spec = CLISignatureSpec(
            ignore_on_cmd_line=cli_attrs.ignore_on_cmd_line(func),
            ignore_in_config=cli_attrs.ignore_in_config(func),
            parse_config_as_cli=cli_attrs.parse_config_as_cli(func),
            typecheck=cli_attrs.typecheck(func),
            parse_env=cli_attrs.parse_env(func),
            metavars=cli_attrs.metavars(func),
            named_groups=cli_attrs.named_groups(func),
            parse_order=cli_attrs.parse_order(func),
            require_options=cli_attrs.require_options(func),
        )

        return (
            spec
            if constructor is None
            else spec.overriding(CLISignatureSpec.from_callable(constructor))
        )

    @property
    def nonnull_attrs(self):
        attrs = ((k, getattr(self, k)) for k in self._fields)
        return dict((k, v) for k, v in attrs if v is not None)

    def overriding(self, *others: "CLISignatureSpec") -> "CLISignatureSpec":
        namespaces = [n.nonnull_attrs for n in chain((self,), others)]
        metavars = ChainMap(*filter(None, (n.get("metavars") for n in namespaces)))
        attrs = ChainMap({"metavars": metavars}, *namespaces)
        if attrs.get("require_options") is None:
            attrs["require_options"] = False
        return CLISignatureSpec(**attrs)

    def configure(
        self, sig: Signature, warn_missing: Optional[Callable[[str], None]] = None
    ) -> "FinalCLISignatureSpec":
        params = sig.parameters
        all_names = set(params)

        if self.parse_env is None:
            parse_env = {}
        else:
            parse_env = {
                name: envname
                for name, envname in self.parse_env.items()
                if name in all_names
            }
            if len(parse_env) < len(self.parse_env):
                raise NameError(
                    "None of the names {} in parse_env occurred in signature {}".format(
                        tuple(name for name in parse_env if name not in all_names), sig
                    )
                )

        if self.metavars is None:
            metavars = {}
        else:
            metavars = {
                name: metavar
                for name, metavar in self.metavars.items()
                if name in all_names
            }
            if warn_missing is not None and len(metavars) < len(self.metavars):
                warn_missing(
                    "None of names {} in metavars occurred in signature {}".format(
                        tuple(name for name in metavars if name not in all_names), sig
                    )
                )

        if self.named_groups is None:
            named_groups = {}
        else:
            named_groups = {}
            memo = set()
            prior = []
            for name, group in self.named_groups.items():
                prior.append(name)
                group = set(group)

                if warn_missing is not None:
                    repeated = group.intersection(memo)
                    if repeated:
                        warn_missing(
                            "Names {} in named group '{}' occurred in at least one of prior named groups {}".format(
                                tuple(repeated), name, prior
                            )
                        )
                    missing = group.difference(all_names)
                    if missing:
                        warn_missing(
                            "Names {} in named group '{}' are not present in signature {}".format(
                                tuple(missing), name, sig
                            )
                        )

                named_groups[name] = group.difference(memo).intersection(all_names)
                memo.update(group)

        if self.parse_order is None:
            parse_order = list(params)
        else:
            parse_order = compute_parse_order(self.parse_order, all_names)

        def param_names(attr, invert=False):
            names = filter_params(
                getattr(self, attr), params, argname=attr, warn_missing=warn_missing
            )
            return all_names.difference(names) if invert else set(names)

        return FinalCLISignatureSpec(
            parse_cmd_line=param_names("ignore_on_cmd_line", invert=True),
            parse_config=param_names("ignore_in_config", invert=True),
            parse_config_as_cli=param_names("parse_config_as_cli", invert=False),
            typecheck=param_names("typecheck", invert=False),
            parse_env=parse_env,
            parse_order=parse_order,
            metavars=metavars,
            named_groups=named_groups,
            all_names=all_names,
            positional_names=list(leading_positionals(sig.parameters, names_only=True)),
            require_options=self.require_options,
        )


class FinalCLISignatureSpec(NamedTuple):
    parse_cmd_line: Set[str]
    parse_config: Set[str]
    parse_config_as_cli: Set[str]
    parse_env: Mapping[str, str]
    parse_order: List[str]
    typecheck: Set[str]
    metavars: Mapping[str, str]
    named_groups: Mapping[str, Set[str]]
    all_names: Set[str]
    positional_names: List[str]
    require_options: bool

    @property
    def parsed(self) -> Set[str]:
        return self.parse_cmd_line.union(self.parse_config).union(self.parse_env)

    @property
    def have_fallback(self) -> Set[str]:
        return self.parse_config.union(self.parse_env)


class UnknownArgSpecifier(TypeError):
    def __init__(self, spec, argname: Optional[str]):
        super().__init__(spec, argname)

    def __str__(self):
        spec = self.args[0]
        msg = "Can't filter function params with specifier {} of type {}".format(
            repr(spec), type(spec)
        )
        if self.args[1]:
            msg += "; occurred for attribute '{}'".format(self.args[1])
        return "{}({})".format(type(self).__name__, repr(msg))


@singledispatch
def filter_params(
    spec: AnyArgNameSpec,
    params: Mapping[str, Parameter],
    argname: Optional[str] = None,
    warn_missing: Optional[Callable[[str], None]] = None,
) -> Iterable[str]:
    if spec is None:
        return ()
    if isinstance(spec, Collection):
        return chain.from_iterable(
            filter_params(s, params, warn_missing=warn_missing) for s in spec
        )
    else:
        raise UnknownArgSpecifier(spec, argname)


@filter_params.register(bool)
def _filter_params_bool(spec: bool, params, argname=None, warn_missing=None):
    return iter(params) if spec else ()


@filter_params.register(str)
def _filter_params_str(spec: str, params, argname=None, warn_missing=None):
    if spec.isidentifier():
        if spec in params:
            return (spec,)
        else:
            if warn_missing is not None:
                warn_missing(
                    "Name {} is not in signature {}{}".format(
                        repr(spec),
                        Signature(params.values()),
                        "; occurred for '{}'".format(argname) if argname else "",
                    )
                )
            return ()
    else:
        return _filter_params_re(re.compile(spec), params, argname)


@filter_params.register(type(re.compile("")))
def _filter_params_re(spec: Pattern, params, argname=None, warn_missing=None):
    names = filter(spec.fullmatch, params)
    if warn_missing is not None:
        names = list(names)
        if not names:
            warn_missing(
                "No names in signature {} match pattern {}{}".format(
                    Signature(params.values()),
                    repr(spec.pattern),
                    "; occurred for '{}'".format(argname) if argname else "",
                )
            )
    return names


@filter_params.register(_ParameterKind)
def _filter_params_kind(spec: _ParameterKind, params, argname=None, warn_missing=None):
    return (name for name, param in params.items() if param.kind == spec)


@filter_params.register(ArgKind)
def _filter_params_argkind(spec: ArgKind, params, argname=None, warn_missing=None):
    kinds = spec.value
    return (name for name, param in params.items() if param.kind in kinds)


def compute_parse_order(
    parse_order: List[Union[str, type(Ellipsis)]], param_names: Set[str]
):
    if parse_order is None:
        return list(param_names)
    parse_order = _validate_parse_order(*parse_order)

    if any((n not in param_names) or (n is not Ellipsis) for n in parse_order):
        raise NameError(
            "parse_order entries must all be in {}; got {}".format(
                param_names, parse_order
            )
        )

    try:
        ellipsis_ix = parse_order.index(Ellipsis)
    except ValueError:
        head, tail = parse_order, ()
        middle = ()
    else:
        head, tail = parse_order[:ellipsis_ix], parse_order[ellipsis_ix + 1 :]
        middle = param_names.difference(head).difference(tail)

    return [
        *(h for h in head if h in param_names),
        *middle,
        *(t for t in tail if t in param_names),
    ]



# coding:utf-8
import typing
import enum
import itertools
import numbers
import pathlib
from bourbaki.introspection.types import (
    get_constraints,
    get_bound,
    get_generic_args,
    reconstruct_generic,
    issubclass_generic,
    is_top_type,
    is_named_tuple_class,
    NonAnyStrCollection,
)
from bourbaki.introspection.classes import parameterized_classpath
from bourbaki.introspection.generic_dispatch import GenericTypeLevelSingleDispatch
from bourbaki.application.completion.completers import (
    completer_argparser_from_type,
    CompletePythonClasses,
    CompletePythonSubclasses,
    CompletePythonCallables,
    CompleteFiles,
    CompleteFilesAndDirs,
    CompleteChoices,
    CompleteEnum,
    CompleteUnion,
    CompleteTuple,
    CompleteKeyValues,
    CompleteFloats,
    CompleteInts,
    CompleteBools,
    NoComplete,
)
from .cli_repr_ import cli_repr
from .utils import File

NoneType = type(None)


cli_completer = GenericTypeLevelSingleDispatch(
    "cli_completer", isolated_bases=[typing.Union]
)

cli_completer.register_all(int, bytes, bytearray, numbers.Integral, as_const=True)(
    CompleteInts()
)

cli_completer.register_all(float, numbers.Real, as_const=True)(CompleteFloats())

cli_completer.register(bool, as_const=True)(CompleteBools())

cli_completer.register(pathlib.Path, as_const=True)(CompleteFilesAndDirs())

cli_completer.register(File, as_const=True)(CompleteFiles())

cli_completer.register(typing.Callable, as_const=True)(CompletePythonCallables())

cli_completer.register(NoneType, as_const=True)(NoComplete)

cli_completer.register_all(enum.Enum, enum.Flag, enum.IntEnum, enum.IntFlag)(
    CompleteEnum
)


@cli_completer.register(NonAnyStrCollection)
def completer_for_collection(coll, v=None):
    if v is None:
        return
    return cli_completer(v)


@cli_completer.register(typing.Mapping)
def completer_for_mapping(m, k=None, v=None):
    return CompleteKeyValues(cli_completer(k), cli_completer(v))


@cli_completer.register(typing.Type)
def completer_for_type(t, supertype=None):
    if supertype is None or is_top_type(supertype):
        return CompletePythonClasses()
    elif isinstance(supertype, typing.TypeVar):
        bound = get_bound(supertype)
        if bound:
            supertype = bound
        else:
            constraints = get_constraints(supertype)
            if constraints:
                return CompleteChoices(*map(parameterized_classpath, constraints))
            return CompletePythonClasses()

    return CompletePythonSubclasses(supertype)


@cli_completer.register(typing.Tuple)
def completer_for_tuple(t, *contents):
    if not contents:
        if is_named_tuple_class(t):
            contents = list(t.__annotations__.items())
            return _multi_completer(CompleteTuple, *contents)
        return completer_for_collection(t)
    elif contents[-1] is Ellipsis:
        return completer_for_collection(t, *contents[:-1])
    else:
        return _multi_completer(CompleteTuple, *contents)


@cli_completer.register(typing.Union)
def completer_for_union(u, *types):
    types = [t for t in types if t not in (NoneType, None)]
    if all(issubclass_generic(t, typing.Tuple) for t in types):
        # complete a Union of tuples as a tuple of unions - a fair approximation that parses
        argtups = (get_generic_args(t) for t in types)
        unionargtups = itertools.zip_longest(*argtups, fillvalue=NoneType)
        uniontypes = (typing.Union[args] for args in unionargtups)
        return _multi_completer(CompleteTuple, *uniontypes)
    return _multi_completer(CompleteUnion, *types, remove_no_complete=True)


# by default return a no-op completer but don't raise a NotImplementedError; this lets the caller perform a best-effort
# completion improvement in a principled way, as in a case where a NamedTuple class has no type-based completer
# available for one of its entries, in which case a name-based completer can be created for that position, which at
# least supplies the user with an informative string as to the meaning of the value that should be passed there
cli_completer.register(typing.Any, as_const=True)(NoComplete)


def _multi_completer(completer_cls, *types_or_tups, remove_no_complete=False):
    completers = []
    for t in types_or_tups:
        if isinstance(t, tuple):
            name, t = t
        else:
            name = None

        nocomplete_types = (type(NoComplete), NoneType)
        try:
            comp = cli_completer(t)
        except (NotImplementedError, TypeError):
            comp = NoComplete

        if isinstance(comp, nocomplete_types):
            comp = NoComplete if name is None else CompleteChoices(name.upper())

        if not remove_no_complete or not isinstance(comp, nocomplete_types):
            completers.append(comp or NoComplete)

    return completer_cls(*completers) if completers else None


# register this as a helper so that manually-defined ArgumentParsers may still have meaningful completions installed
completer_argparser_from_type.register(type)(cli_completer)



# coding:utf-8
import typing
import decimal
import enum
import fractions
import pathlib
import datetime
import ipaddress
import collections
from urllib.parse import ParseResult as URL, urlparse
import uuid
from inspect import Parameter
from functools import lru_cache
from warnings import warn
from bourbaki.introspection.types import (
    get_named_tuple_arg_types,
    is_named_tuple_class,
    issubclass_generic,
    is_top_type,
    LazyType,
    NonStrCollection,
)
from bourbaki.introspection.callables import signature
from bourbaki.introspection.generic_dispatch import (
    GenericTypeLevelSingleDispatch,
    UnknownSignature,
    DEBUG,
)
from bourbaki.introspection.generic_dispatch_helpers import (
    CollectionWrapper,
    MappingWrapper,
    UnionWrapper,
    TupleWrapper,
    LazyWrapper,
)
from .cli_complete import cli_completer
from .cli_nargs_ import check_tuple_nargs, check_union_nargs, cli_nargs
from .cli_repr_ import cli_repr
from .exceptions import (
    CLITypedInputError,
    CLIIOUndefined,
    CLIUnionInputError,
    CLINestedCollectionsNotAllowed,
)
from .parsers import (
    parse_iso_date,
    parse_iso_datetime,
    parse_range,
    parse_regex,
    parse_regex_bytes,
    parse_bool,
    parse_path,
    EnumParser,
    FlagParser,
    TypeCheckImportType,
    TypeCheckImportFunc,
)
from .utils import (
    File,
    Empty,
    identity,
    KEY_VAL_JOIN_CHAR,
    parser_constructor_for_collection,
)

NoneType = type(None)


class InvalidCLIParser(TypeError):
    pass


def cli_split_keyval(s: str):
    i = s.index(KEY_VAL_JOIN_CHAR)
    return s[:i], s[i + 1 :]


def cli_parse_bytes(seq: typing.Sequence[str]):
    return bytes(map(int, seq))


def cli_parse_bytearray(seq: typing.Sequence[str]):
    return bytearray(map(int, seq))


def cli_parse_bool(s: typing.Union[str, bool]):
    # have to allow bools for the flag case; argparse includes an implicit boolean default in that case
    if isinstance(s, bool):
        return s
    return parse_bool(s, exc_type=CLITypedInputError)


cli_parse_methods = {
    int: int,
    bool: cli_parse_bool,
    float: float,
    str: str,
    complex: complex,
    decimal.Decimal: decimal.Decimal,
    fractions.Fraction: fractions.Fraction,
    bytes: cli_parse_bytes,
    bytearray: cli_parse_bytearray,
    range: parse_range,
    datetime.date: parse_iso_date,
    datetime.datetime: parse_iso_datetime,
    pathlib.Path: parse_path,
    typing.Pattern: parse_regex,
    typing.Pattern[bytes]: parse_regex_bytes,
    uuid.UUID: uuid.UUID,
    URL: urlparse,
    ipaddress.IPv4Address: ipaddress.IPv4Address,
    ipaddress.IPv6Address: ipaddress.IPv6Address,
    Empty: identity,
}


@lru_cache(None)
def _validate_parser(func):
    try:
        sig = signature(func)
    except ValueError:
        annotation = None
    else:
        msg = "CLI parsers must have only a single required positional arg; "
        if len(sig.parameters) == 0:
            raise InvalidCLIParser(msg + "{} takes no arguments!".format(func))

        required = [
            p
            for p in sig.parameters.values()
            if p.default is Parameter.empty
            and p.kind not in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)
        ]

        if len(required) > 1:
            raise InvalidCLIParser(
                msg
                + "{} requires {} args with signature {}".format(
                    func, len(required), sig
                )
            )
        elif any(p.kind == Parameter.KEYWORD_ONLY for p in required):
            raise InvalidCLIParser(
                msg + "{} requires keyword args with signature {}".format(func, sig)
            )
        else:
            param = next(iter(sig.parameters.values()))
            if param.kind in (Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD):
                raise InvalidCLIParser(
                    msg
                    + "{} takes only keyword args with signature {}".format(func, sig)
                )
            if param.kind is Parameter.VAR_POSITIONAL:
                warn(
                    "{} accepts *args for its first argument; when called as a parser, only one arg will be passed. "
                    "Is this what was intended?"
                )

        annotation = param.annotation

    return func, annotation


class CLIParserDispatch(GenericTypeLevelSingleDispatch):
    def register(
        self,
        *sig,
        debug: bool = DEBUG,
        as_const: bool = False,
        derive_nargs: bool = False,
        derive_repr: bool = False,
        derive_completer: bool = False
    ):
        """
        Register a parser for a type, and optionally, if the following are defined for the parser's single arg
        annotation type and not yet registered, register them too for the type:
        - cli_nargs
        - cli_repr
        - cli_completer
        when derive_nargs, derive_repr, derive_completer respectively are True.
        """
        dec = super().register(*sig, debug=debug, as_const=as_const)
        if not as_const:
            return dec

        def maybe_register_nargs_repr_completer(f):
            # register an inferred nargs with cli_nargs if possible
            func, type_ = _validate_parser(f)
            if type_ is not None and type_ is not Parameter.empty:
                for derive, dispatcher in [
                    (derive_nargs, cli_nargs),
                    (derive_repr, cli_repr),
                    (derive_completer, cli_completer),
                ]:
                    # if derivation is indicated and the signature is not already registered
                    if derive and sig not in dispatcher.funcs:
                        try:
                            value = dispatcher(type_)
                        except NotImplementedError:
                            pass
                        else:
                            dispatcher.register(*sig, debug=debug, as_const=True)(value)

            return dec(f)

        return maybe_register_nargs_repr_completer


cli_parser = CLIParserDispatch(__name__, isolated_bases=[typing.Union])


@cli_parser.register(typing.Any)
def cli_parser_any(t, *args):
    if is_top_type(t):
        return identity
    raise CLIIOUndefined((t, *args))


@cli_parser.register(typing.Callable)
class TypeCheckImportFuncCLI(TypeCheckImportFunc):
    exc_cls = CLITypedInputError


@cli_parser.register(typing.Type)
class TypeCheckImportTypeCLI(TypeCheckImportType):
    exc_cls = CLITypedInputError


class GenericCLIParserMixin:
    getter = cli_parser
    get_reducer = staticmethod(parser_constructor_for_collection)


@cli_parser.register(typing.Collection)
class CollectionCLIParser(GenericCLIParserMixin, CollectionWrapper):
    def __init__(self, coll_type, val_type=object):
        if issubclass_generic(val_type, NonStrCollection):
            raise CLINestedCollectionsNotAllowed((coll_type, val_type))
        super().__init__(coll_type, val_type)


@cli_parser.register(typing.Collection[NonStrCollection])
class NestedCollectionCLIParser(GenericCLIParserMixin, CollectionWrapper):
    def __init__(self, coll_type, val_type=object):
        super().__init__(coll_type, val_type)


@cli_parser.register(typing.Mapping)
class MappingCLIParser(GenericCLIParserMixin, MappingWrapper):
    constructor_allows_iterable = True

    def __init__(self, coll_type, key_type=object, val_type=object):
        if issubclass_generic(key_type, NonStrCollection) or issubclass_generic(
            val_type, NonStrCollection
        ):
            raise CLINestedCollectionsNotAllowed((coll_type, key_type, val_type))
        super().__init__(coll_type, key_type, val_type)
        coll_type = self.reduce
        if issubclass(coll_type, (collections.Counter, collections.ChainMap)):
            self.constructor_allows_iterable = False

    def __call__(self, args):
        keyvals = map(cli_split_keyval, args)
        if not self.constructor_allows_iterable:
            return self.reduce(dict(self.call_iter(keyvals)))
        return super().__call__(keyvals)


@cli_parser.register(typing.Union)
class UnionCLIParser(GenericCLIParserMixin, UnionWrapper):
    reduce = staticmethod(next)
    tolerate_errors = (CLIIOUndefined, UnknownSignature)
    exc_class = CLIUnionInputError

    def __init__(self, u, *types):
        check_union_nargs(*types)
        super().__init__(u, *types)

    def __call__(self, arg):
        if arg is None and self.is_optional:
            return arg
        return super().__call__(arg)


@cli_parser.register(typing.Tuple)
class TupleCLIParser(GenericCLIParserMixin, TupleWrapper):
    _collection_cls = CollectionCLIParser
    _nargs = None
    _entry_nargs = None
    require_same_len = False

    def __new__(cls, t, *types):
        if is_named_tuple_class(t):
            types = get_named_tuple_arg_types(t)

        self = TupleWrapper.__new__(cls, t, *types)
        if Ellipsis in types:
            return self

        self._entry_nargs, self._nargs = check_tuple_nargs(t, *types)
        self.require_same_len = all(n in (None, 1) for n in self._entry_nargs)
        return self

    def iter_chunks(self, args):
        ix = 0
        for n in self._entry_nargs:
            if n is None:
                yield args[ix]
                ix += 1
            elif isinstance(n, int):
                yield args[ix : ix + n]
                ix += n
            else:
                yield args[ix:]
                ix = None

    def call_iter(self, arg):
        return (f(a) for f, a in zip(self.funcs, self.iter_chunks(arg)))


@cli_parser.register(LazyType)
class LazyCLIParser(GenericCLIParserMixin, LazyWrapper):
    pass


@cli_parser.register_all(enum.Enum, enum.IntEnum)
def cli_enum_parser(enum_):
    return EnumParser(enum_).cli_parse


@cli_parser.register_all(enum.Flag, enum.IntFlag)
def cli_flag_parser(enum_):
    return FlagParser(enum_).cli_parse


# File subclasses are their own parsers
cli_parser.register(File)(identity)


cli_parser.register_from_mapping(cli_parse_methods, as_const=True)


cli_option_parser = GenericTypeLevelSingleDispatch(
    "cli_option_parser", isolated_bases=cli_parser.isolated_bases
)


@cli_option_parser.register(typing.Collection[NonStrCollection])
class CollectionCLIOptionParser(GenericCLIParserMixin, CollectionWrapper):
    pass
